"""Classes and methods to compute the gradient of functions involving jax arrays transformations"""

from __future__ import annotations

from typing import Any, Generic, Hashable, Protocol, Tuple, TypeVar

import jax_dataclasses as jdc
import optax

from jax_utils.compilation import jit_when_compilation_enabled
from jax_utils.dynamics import JaxDynamics
from jax_utils.jax_tensor import AverageableJaxArrayLike
from jax_utils.vectorization import JaxDataclassNestedConvertibleToAxes, vectorize

State = TypeVar("State")
State_contra = TypeVar("State_contra", contravariant=True)
Action = TypeVar("Action")
Observation_co = TypeVar("Observation_co", covariant=True)
Cost_co = TypeVar("Cost_co", covariant=True, bound=AverageableJaxArrayLike)
OptimizerState = TypeVar("OptimizerState", bound=optax.OptState)
GradientUpdates = TypeVar("GradientUpdates", bound=optax.Updates)
AxisType = TypeVar("AxisType", bound=Hashable)
AxisType_contra = TypeVar("AxisType_contra", contravariant=True, bound=Hashable)


class BaseGradientStep(
    JaxDataclassNestedConvertibleToAxes[AxisType_contra],
    Protocol[AxisType_contra, State_contra, Action, Cost_co, OptimizerState],
):
    """Interface for gradient steps on the  :class:`jax_utils.markov_decision_process.Dynamics` cost of a Markov
    Decision Process."""

    # pylint: disable=C0116
    def init_optimizer(self, init_action: Action) -> OptimizerState: ...

    # pylint: disable=C0116
    def compute_gradient(
        self,
        state: State_contra,
        action: Action,
    ) -> Tuple[Cost_co, Action]: ...

    # pylint: disable=C0116
    def update(
        self, action: Action, gradient_value: GradientUpdates, opt_state: OptimizerState
    ) -> Tuple[Action, OptimizerState]: ...

    def __call__(
        self,
        state: State_contra,
        action: Action,
        opt_state: OptimizerState,
    ) -> Tuple[Action, OptimizerState, Cost_co]: ...


@jdc.pytree_dataclass(frozen=True)
class GradientStep(
    BaseGradientStep[AxisType, State, Action, Cost_co, OptimizerState],
    Generic[AxisType, State, Action, Cost_co, Observation_co, OptimizerState],
):
    """The gradient step on the :class:`jax_utils.markov_decision_process.Dynamics` cos
    of a Markov Decision Process.

    Args:
        optimizer (jdc.Static[optax.GradientTransformation]): an optimizer (defining the
            stochastic gradient descent variant: RMSProp, Adam, etc..)
        dynamics (JaxDynamics[Any, State, Action, Cost_co, Observation_co]): an MDP dynamics
    """

    optimizer: jdc.Static[optax.GradientTransformation]
    dynamics: JaxDynamics[Any, State, Action, Cost_co, Observation_co]

    @jit_when_compilation_enabled()
    def init_optimizer(self, init_action: Action) -> OptimizerState:
        """Initialize optimizer

        Args:
            init_action (Action): initial action before optimization starts

        Returns:
            OptimizerState: initialized optimizer state
        """
        return self.optimizer.init(init_action)

    @jit_when_compilation_enabled()
    def compute_gradient(
        self,
        state: State,
        action: Action,
    ) -> Tuple[Cost_co, Action]:
        """Given a state-action pair of the MDP, computes the cost and gradient
        of the cost of a ``dynamics`` w.r.t. the action.

        Args:
            state (State): state of the MDP
            action (Action): action of the MDP

        Returns:
            Tuple[Cost_co, Action]: cost and gradient of the action
        """
        return self.dynamics.compute_gradient(
            state=state,
            action=action,
        )

    @jit_when_compilation_enabled()
    def update(
        self, action: Action, gradient_value: GradientUpdates, opt_state: OptimizerState
    ) -> Tuple[Action, OptimizerState]:
        """Update the action and optimizer state

        Args:
            action (Action): action of the MDP
            gradient_value (GradientUpdates): gradient of the action
            opt_state (OptimizerState): optimizer state

        Returns:
            Tuple[Action, OptimizerState]: the new action and optimizer state
        """
        updates, updated_opt_state = self.optimizer.update(gradient_value, opt_state, action)
        updated_action = optax.apply_updates(action, updates)
        return updated_action, updated_opt_state

    @jit_when_compilation_enabled()
    def __call__(
        self,
        state: State,
        action: Action,
        opt_state: OptimizerState,
    ) -> Tuple[Action, OptimizerState, Cost_co]:
        """Computes the gradient of the cost w.r.t. the action and updates the action and optimizer state.

        Args:
            state (State): state of the MDP
            action (Action): action of the MDP
            opt_state (OptimizerState): optimizer state

        Returns:
            Tuple[Action, OptimizerState, Cost_co]: a new action, optimizer state and the value of the cost
                associated to the previous action
        """
        cost_value, gradient_value = self.compute_gradient(
            state=state,
            action=action,
        )
        updated_action, updated_opt_state = self.update(action, gradient_value, opt_state)
        return updated_action, updated_opt_state, cost_value


@jdc.pytree_dataclass(frozen=True)
class VectorizedGradientStep(BaseGradientStep[AxisType, State, Action, Cost_co, OptimizerState]):
    """A gradient step that is `vectorized <https://jax.readthedocs.io/en/latest/jax-101/03-vectorization.html>`_ on a
    specific axis

    Args:
        gradient_step (BaseGradientStep[AxisType, State, Action, Cost_co, OptimizerState]): the gradient step to be
            vectorized
        vectorized_axis (jdc.Static[AxisType]): "named" axis to be vectorized
    """

    gradient_step: BaseGradientStep[AxisType, State, Action, Cost_co, OptimizerState]
    vectorized_axis: jdc.Static[AxisType]

    @jit_when_compilation_enabled()
    def init_optimizer(self, init_action: Action) -> OptimizerState:
        return self.gradient_step.init_optimizer(init_action)

    @vectorize()
    def compute_gradient(
        self,
        state: State,
        action: Action,
    ) -> Tuple[Cost_co, Action]:
        return self.gradient_step.compute_gradient(
            state,
            action,
        )

    @jit_when_compilation_enabled()
    def update(
        self, action: Action, gradient_value: GradientUpdates, opt_state: OptimizerState
    ) -> Tuple[Action, OptimizerState]:
        return self.gradient_step.update(action, gradient_value, opt_state)

    @jit_when_compilation_enabled()
    def __call__(
        self,
        state: State,
        action: Action,
        opt_state: OptimizerState,
    ) -> Tuple[Action, OptimizerState, Cost_co]:
        cost_value, gradient_value = self.compute_gradient(state, action)
        updated_action, updated_opt_state = self.update(action, gradient_value, opt_state)
        return updated_action, updated_opt_state, cost_value



"""Classes and methods to easily map pytrees to axes (e.g., for vectorization)"""

from typing import Any, Hashable, Optional, Protocol, TypeVar, runtime_checkable

from jax.tree_util import tree_map
from typing_extensions import Self

AxisType = TypeVar("AxisType", bound=Hashable)
AxisType_contra = TypeVar("AxisType_contra", contravariant=True)


# pylint: disable=C0115
@runtime_checkable
class ConvertibleToAxes(Protocol[AxisType_contra]):
    def convert_to_axes(self, axis: Optional[AxisType_contra]) -> Self:
        """
        Returns an object that can be used in argument ``in_axes`` or ``out_axes`` of
        `jax.vmap <https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html>`_
        or `jax.pmap <https://jax.readthedocs.io/en/latest/_autosummary/jax.pmap.html>`_

        Args:
            axis (Optional[AxisType_contra]): a "named" axis over which to apply
                `vectorization <https://jax.readthedocs.io/en/latest/jax-101/03-vectorization.html>`_.

        Returns:
            Self: same object as ``self`` but with all array-like objects replaced by axes
        """


def pytree_to_axes(
    pytree: Any, vectorized_axis: AxisType, default_axis: Optional[int] = None
) -> Any:
    """
    Transform all the :class:`ConvertibleToAxes` leafs of a given
    `pytree <https://jax.readthedocs.io/en/latest/pytrees.html>`_ to axes by applying method
    :meth:`ConvertibleToAxes.convert_to_axes`. This is useful for
    `vectorizing <https://jax.readthedocs.io/en/latest/jax-101/03-vectorization.html>`_ functions
    involving :class:`ConvertibleToAxes` objects.

    Args:
        pytree (Any): any Python pytree containing :class:`ConvertibleToAxes` leafs
        vectorized_axis (AxisType): a "named" axis over which to apply
            `vectorization <https://jax.readthedocs.io/en/latest/jax-101/03-vectorization.html>`_
        default_axis (Optional[int], optional): A default axis value for leafs that are not :class:`ConvertibleToAxes`.
            Defaults to None.

    Returns:
        Any: same pytree as given in input but where all :class:`ConvertibleToAxes` are converted to axes.
    """
    return tree_map(
        lambda x: x.convert_to_axes(vectorized_axis)
        if isinstance(x, ConvertibleToAxes)
        else default_axis,
        pytree,
        is_leaf=lambda x: isinstance(x, ConvertibleToAxes),
    )



"""Classes and methods to to jit-compile functions involving jax arrays transformations"""

from functools import wraps
from typing import Callable, Protocol

from jax import jit
from typing_extensions import Self


class JaxCompilableProtocol(Protocol):
    """All classes implementing this interface should implement property ``is_compilation_enabled``
    indicating whether the methods involving JAX arrays should be
    `jit-compiled <https://jax.readthedocs.io/en/latest/jax-101/02-jitting.html>`_.
    """

    # pylint: disable=C0116
    @property
    def is_compilation_enabled(self) -> bool:
        pass


class BaseJaxCompilable(JaxCompilableProtocol, Protocol):
    """Subclassing ``BaseCompilableJax`` allows to easily enable/disable
    `jit-compilation <https://jax.readthedocs.io/en/latest/jax-101/02-jitting.html>`_
    of methods involving JAX arrays.

    Use ``with_optional_jax_jit`` decorator to compile a method only when
    ``is_compilation_enabled`` is ``True`` (``False`` by default).

    To enable (resp. disable) jit-compilation, one only needs to call method ``enable_compilation``
    (resp. ``disable_compilation``). By default, jit-compilation is disabled.
    """

    @property
    def is_compilation_enabled(self) -> bool:
        return hasattr(self, "_is_compilation_enabled") and getattr(self, "_is_compilation_enabled")

    # pylint: disable=C0116
    def enable_compilation(self) -> Self:
        object.__setattr__(self, "_is_compilation_enabled", True)
        return self

    # pylint: disable=C0116
    def disable_compilation(self) -> Self:
        object.__setattr__(self, "_is_compilation_enabled", False)
        return self


def jit_when_compilation_enabled(**jax_jit_args) -> Callable[[Callable], Callable]:
    """Parametrized decorator for methods of classes implementing ``CompilableJaxProtocol`` interface.
    Allows to `jit-compile <https://jax.readthedocs.io/en/latest/jax-101/02-jitting.html>`_
    some methods only when compilation is enabled.

    Returns:
        Callable[[Callable], Callable]: decorator with parameters ``jax_jit_args`` specified
    """

    def decorator(
        func: Callable,
    ) -> Callable:
        @wraps(func)
        def wrapper(self: JaxCompilableProtocol, *args, **kwargs):
            if self.is_compilation_enabled:
                return jit(func, **jax_jit_args)(self, *args, **kwargs)
            return func(self, *args, **kwargs)

        return wrapper

    return decorator



"""High-level abstractions for decision problems (Markov Decision processes, etc...)"""

from typing import Protocol, Tuple, TypeVar

State = TypeVar("State")
State_contra = TypeVar("State_contra", contravariant=True)
Action = TypeVar("Action")
Action_contra = TypeVar("Action_contra", contravariant=True)
Observation = TypeVar("Observation")
Observation_co = TypeVar("Observation_co", covariant=True)
Observation_contra = TypeVar("Observation_contra", contravariant=True)
Cost = TypeVar("Cost")
Cost_co = TypeVar("Cost_co", covariant=True)
Cost_contra = TypeVar("Cost_contra", contravariant=True)
RegularizedCost = TypeVar("RegularizedCost")


class Dynamics(Protocol[State, Action_contra, Cost_co, Observation_co]):
    """Interface defining the dynamics of a
    `(Partially Observable) <https://en.wikipedia.org/wiki/Partially_observable_Markov_decision_process>`_
    `Markov Decision Process <https://en.wikipedia.org/wiki/Markov_decision_process>`_.

    When an "agent" interacting with the (PO)MDP plays an "action" (a.k.a. "control")
    in a given "state", the (PO)MDP transitions to a new state and the agent observes
    some signal/feedback in the form of a "cost"/"reward" as well as additional "observations".

    A `Dynamics` is therefore a callable that maps a state-action pair to a state-cost-observation tuple.
    """

    def __call__(
        self, state: State, action: Action_contra
    ) -> Tuple[State, Cost_co, Observation_co]:
        pass


class CostRegularizer(
    Protocol[State_contra, Action_contra, Cost_contra, Observation_contra, Cost_co]
):
    """Interface for callables that map any state-action-cost-observation tuple to a new "regularized" cost.

    `More about regularization <https://en.wikipedia.org/wiki/Regularization_(mathematics)>`_.

    Example: one may want to penalize `action` with high norms, etc...
    """

    def __call__(
        self,
        state: State_contra,
        action: Action_contra,
        cost: Cost_contra,
        observation: Observation_contra,
    ) -> Cost_co:
        pass


class RegularizedDynamics(
    Dynamics[State, Action, RegularizedCost, Observation],
    Protocol[State, Action, Cost, RegularizedCost, Observation],
):
    """Interface defining a wrapper around class :class:`jax_utils.markov_decision_process.Dynamics` that allows to add
    a regularization to the cost.

    A ``RegularizedDynamics`` is itself a :class:`jax_utils.markov_decision_process.Dynamics``.

    `More about regularization <https://en.wikipedia.org/wiki/Regularization_(mathematics)>`_.

    Args:
        dynamics (Dynamics[State, Action, Cost, Observation]): callable defining the dynamics of a (PO)MDP
        cost_regularizer (CostRegularizer[State, Action, Cost, Observation, RegularizedCost]): callable
        defining a cost transformation (regularization)
    """

    dynamics: Dynamics[State, Action, Cost, Observation]
    cost_regularizer: CostRegularizer[State, Action, Cost, Observation, RegularizedCost]

    def __call__(self, state: State, action: Action) -> Tuple[State, RegularizedCost, Observation]:
        state, cost, observation = self.dynamics(state, action)
        regularized_cost = self.cost_regularizer(state, action, cost, observation)
        return state, regularized_cost, observation



"""Wrappers around numpy/jax arrays"""

from __future__ import annotations

from numbers import Number
from typing import (
    AbstractSet,
    Any,
    ClassVar,
    Dict,
    Hashable,
    Iterable,
    Optional,
    Protocol,
    Tuple,
    TypeAlias,
    TypeVar,
    Union,
    runtime_checkable,
)

import jax.numpy as jnp
import numpy as np
from ordered_set import OrderedSet, OrderedSetInitializer
from typing_extensions import Self

from jax_utils.typing import DataclassInstance

Array: TypeAlias = Union[jnp.ndarray, np.ndarray]
Scalar: TypeAlias = Union[float, int]


def check_ndim_in(array: Array, allowed_ndims: Iterable[int]):
    """Checks if the number of dimensions matches some allowed values.

    Args:
        array (Array): numpy or JAX array
        allowed_ndims (Iterable[int]): number of dimensions allowed for ``array``

    Raises:
        ValueError: when the number of dimensions of ``array`` is not present in ``allowed_shapes``.
    """
    if array.ndim not in allowed_ndims:
        raise ValueError(f"Dimension of delivery should be in {allowed_ndims}")


def is_broadcastable(shape_1: Tuple[int, ...], shape_2: Tuple[int, ...]) -> bool:
    """Whether the shapes of 2 arrays/tensors can be
    `broadcasted <https://numpy.org/doc/stable/user/basics.broadcasting.html>`_.

    Args:
        shape_1 (Tuple[int, ...]): shape of the 1st array
        shape_2 (Tuple[int, ...]): shape of the 2nd array

    Returns:
        bool: ``True`` if and only if the 2 arrays/tensors can be broadcasted, False otherwise.
    """
    for dim1, dim2 in zip(shape_1[::-1], shape_2[::-1]):
        if not (dim1 == 1 or dim2 == 1 or dim1 == dim2):
            return False
    return True


T = TypeVar("T", bound=Hashable)


class TensorAxes(OrderedSet[T]):
    """An ordered set of "named" array/tensor axes.

    The position in the set defines the axis index in the tensor.

    The first ``len(self) - tensor_min_dim`` axes are optional:
    if absent, they are assumed to be "flattened".

    The reason for not making the last ``tensor_min_dim`` axes optional instead of the first
    ``len(self) - tensor_min_dim`` axes is to follow the logic of array broadcasting (first dimension can be omitted).
    """

    def __init__(
        self,
        initial: Optional[OrderedSetInitializer[T]] = None,
        tensor_min_dim: int = 0,
    ):
        super().__init__(initial)  # type: ignore
        if tensor_min_dim > len(self):
            raise ValueError(
                "tensor_min_shape_length should be smaller than the number of specified axes"
            )
        if tensor_min_dim < 0:
            raise ValueError("tensor_min_shape_length should be greater than or equal to 0")
        self.tensor_min_nb_dim = tensor_min_dim

    def reverse_index(self, key: T) -> int:
        """Computes the index of element ``key`` but rather than return a non-negative integer like
        method ``index``, it returns a negative integer e.g., -1 if for the last element,
        -2 for the penultimate element, ...

        Args:
            key (T): any element present in ``self`` (which is an ordered set)

        Returns:
            int: negative integer corresponding to the position of ``key`` in ``self``
                (starting from the last and decrementing by 1 for each element)
        """
        return -(len(self) - self.index(key))

    @property
    def _first_mandatory_idx(self) -> int:
        return -self.tensor_min_nb_dim if self.tensor_min_nb_dim > 0 else len(self)

    @property
    def mandatory(self) -> OrderedSet[T]:
        """
        Returns:
            OrderedSet[T]: ordered set of non-optional axes.
        """
        return OrderedSet(list(self[self._first_mandatory_idx :]))

    @property
    def optional(self) -> OrderedSet[T]:
        """
        Returns:
            OrderedSet[T]: ordered set of optional axes.
        """
        return OrderedSet(list(self[: self._first_mandatory_idx]))

    def __repr__(self) -> str:
        return type(self).__name__ + "\n| ".join(
            [""]
            + [f"{self.reverse_index(axis)} (optional): {axis}" for axis in self.optional]
            + [f"{self.reverse_index(axis)}: {axis}" for axis in self.mandatory]
        )


ArrayType = TypeVar("ArrayType", bound=Array)
AxisType = TypeVar("AxisType", bound=Hashable)


def expand_dims_axis(
    tensor_axes: TensorAxes[AxisType], missing_axes: AbstractSet[AxisType]
) -> Tuple[int, ...]:
    """Compute ``axis`` argument to pass to ``numpy.expand_dims(a, axis)`` function when some axes in ``tensor_axes``
    are not present in the numpy array ``a`` (they are "flattened").

    Args:
        tensor_axes (TensorAxes[AxisType]): some array/tensor axes
        missing_axes (AbstractSet[AxisType]): missing axes in the array/tensor

    Raises:
        ValueError: raises an error when ``missing_axes`` is not a subset of ``tensor_axes``

    Returns:
        Tuple[int, ...]: the ``axis`` argument to pass to ``numpy.expand_dims(a, axis)``
    """
    if not set(missing_axes).issubset(tensor_axes):
        raise ValueError(
            f"The following missing axes are invalid: {', '.join(str(axis) for axis in missing_axes - tensor_axes)}. "
            f"Only the following axes are valid: {', '.join(str(axis) for axis in tensor_axes)}"
        )
    return tuple(
        tensor_axes.reverse_index(axis) for axis in tensor_axes[::-1] if axis in missing_axes
    )


ArrayType_co = TypeVar("ArrayType_co", bound=Array, covariant=True)


class AverageableArrayLike(Protocol[ArrayType_co]):
    """Shared interface of all classes with a ``mean`` method return a scalar array (corresponding to the mean values of
    the initial array).

    Example of classes implementing this interface: (jax) numpy arrays, ...
    """

    def mean(self, *args, **kwargs) -> ArrayType_co:
        """
        Returns:
            ArrayType_co: Should return a scalar array
        """


@runtime_checkable
class Tensor(DataclassInstance, AverageableArrayLike, Protocol[ArrayType, AxisType]):
    """
    A wrapper for numpy/jax arrays with the following additional features:

    - axes are "named" to facilitate manipulation and debugging (an axis "name" could be a string or any other
      hashable ``AxisType``)
    - the returned ``values`` of the tensor can be different from the actual ``array`` given as input at construction,
      this allows to easily implement "change of variables" (a.k.a., "substitution")

    Args:
        _tensor_axes (ClassVar[TensorAxes[AxisType]]): class attribute defining the axes of the ``array``
        array (ArrayType): a (jax) numpy array containing all relevant data

    """

    _tensor_axes: ClassVar[TensorAxes[AxisType]]  # type: ignore[misc]
    array: ArrayType
    # TODO: to facilitate vectorization with vmap, it could be convenient  # pylint: disable=W0511
    # for the user to define an optional `values_axes` attribute (None by default).
    # When not None, `values` property would basically perform a reshape of `array` attribute
    # (permutation of axes). This would allow to put the `vectorized_axis` as the first axis in
    # all input tensors of a vectorized method (see vectorization.py)

    def __post_init__(self):
        self.check_array()

    def check_array(self):
        """Check the validity of the ``array`` attribute at construction.
        To be overriden if needed.
        """
        check_ndim_in(
            self.array, range(self.axes.tensor_min_nb_dim, len(type(self)._tensor_axes) + 1)
        )

    def __getitem__(self, key):
        return type(self)(array=self.array[key])  # type: ignore[call-arg]

    def getitem_from_axes(self, axes_keys: Dict[AxisType, Any]) -> Self:
        """Analogue of method ``__getitem__`` but where array slicing/indexing is explicitly applied to named axis.

        Args:
            axes_keys (Dict[AxisType, Any]): a mapping between axes names and slices/list of indices/...

        Returns:
            Self: a new ``Tensor`` of the same type with restricted data
        """
        array_key = tuple(
            axes_keys[axis] if axis in axes_keys else slice(None, None, None)
            for axis in self.actual_axes
        )
        return self[array_key]

    def __neg__(self) -> Self:
        return self.__class__(array=-self.array)  # type: ignore[call-arg]

    def __abs__(self) -> Self:
        return self.__class__(array=abs(self.array))  # type: ignore[call-arg]

    def __add__(self, other: Union[Scalar, Array, Tensor[ArrayType, AxisType]]) -> Self:
        if isinstance(other, Tensor):
            return self.__class__(array=self.array + other.array)  # type: ignore[call-arg]
        return self.__class__(array=self.array + other)  # type: ignore[call-arg]

    def __sub__(self, other: Union[Scalar, Array, Tensor[ArrayType, AxisType]]) -> Self:
        if isinstance(other, Tensor):
            return self.__class__(array=self.array - other.array)  # type: ignore[call-arg]
        return self.__class__(array=self.array - other)  # type: ignore[call-arg]

    def __mul__(self, other: Union[Scalar, Array, Tensor[ArrayType, AxisType]]) -> Self:
        if isinstance(other, Tensor):
            return self.__class__(array=self.array * other.array)  # type: ignore[call-arg]
        return self.__class__(array=self.array * other)  # type: ignore[call-arg]

    def __truediv__(self, other: Union[Scalar, Array, Tensor[ArrayType, AxisType]]) -> Self:
        if isinstance(other, Tensor):
            return self.__class__(array=self.array / other.array)  # type: ignore[call-arg]
        return self.__class__(array=self.array / other)  # type: ignore[call-arg]

    @classmethod
    def _expand_dims_axis(cls, missing_axes: AbstractSet[AxisType]):
        return expand_dims_axis(cls._tensor_axes, missing_axes)

    @property
    def values(self) -> ArrayType:
        """Override to apply transformations on ``self.array`` ("change of variable").

        For example, if one wants to manipulate (jax) tensors constrained to always have non-negative values in any
        situation (even if the tensor gets updated by some gradient descent procedure for instance), then one can
        override ``values`` by returning ``jax.nn.relu(self.array)``.

        By default ``values`` implements the identity i.e., returns ``self.array``.

        Returns:
            ArrayType: the transformed ``array``
        """
        return self.array

    def reverse_values(self, values: ArrayType) -> ArrayType:
        """Inverse of the transformation implemented in ``values`` property.

        By default, ``self.array = self.values`` so calling ``values`` is equivalent to calling ``array``.

        When the transformation mapping ``array`` to ``values`` is not the identity, ``reverse_values`` should be
        overriden accordingly.

        Args:
            values (ArrayType): the output of property ``values``

        Returns:
            ArrayType: the array obtained by apply the inverse transformation of
                ``values``
        """
        return values

    @property
    def dtype(self) -> np.dtype:
        """Get the type of ``self.values``"""
        return self.values.dtype

    @property
    def shape(self) -> Tuple[int, ...]:
        """Get the shape of ``self.values``"""
        return self.values.shape

    @property
    def ndim(self) -> int:
        """Get the number of dimensions of ``self.values``"""
        return self.values.ndim

    @property
    def axes(self) -> TensorAxes[AxisType]:
        """Possible "named" axes of the tensor"""
        return type(self)._tensor_axes

    @property
    def actual_axes(self) -> TensorAxes[AxisType]:
        """Actual axes of the tensor"""
        return TensorAxes(
            [self.axes[i] for i in range(-len(self.shape), 0)],
            tensor_min_dim=self.axes.tensor_min_nb_dim,
        )

    def mean(self, *args, **kwargs) -> Union[ArrayType, Any]:
        """Compute the mean on ``self.values`` on specified axes"""
        return self.values.mean(*args, **kwargs)

    def mean_over_axes(self, axes: AbstractSet[AxisType]) -> Union[ArrayType, Any]:
        """Averages ``values`` along one or several named axes.

        Args:
            axes (AbstractSet[AxisType]): axes along which the means are computed

        Returns:
            Union[ArrayType, Any]: new array containing the mean ``values``
        """
        axes_to_mean_over = tuple(self.index(axis) for axis in self.actual_axes.intersection(axes))
        return self.mean(axis=axes_to_mean_over)

    def sum(self, *args, **kwargs) -> Union[ArrayType, Any]:
        """Get the sum of the array"""
        return self.values.sum(*args, **kwargs)

    def sum_over_axes(self, axes: AbstractSet[AxisType]) -> Union[ArrayType, Any]:
        """Sums ``values`` along one or several named axes.

        Args:
            axes (AbstractSet[AxisType]): axes along which the sums are computed

        Returns:
            Union[ArrayType, Any]: new array containing the summed ``values``
        """
        axes_to_mean_over = tuple(self.index(axis) for axis in self.actual_axes.intersection(axes))
        return self.sum(axis=axes_to_mean_over)

    def has(self, axis: AxisType) -> bool:
        """Whether ``axis`` is one of the possible axes of the tensor.

        Args:
            axis (AxisType): a "named" axis

        Returns:
            bool: ``True`` if ``axis`` is one of the possible axes of the tensor, ``False`` otherwise.
        """
        return axis in self.axes

    def has_actual(self, axis: AxisType) -> bool:
        """Whether ``axis`` is one of the actual axes of the tensor.

        Args:
            axis (AxisType): a "named" axis

        Returns:
            bool: ``True`` if ``axis`` is one of the actual axes of the tensor, ``False`` otherwise.
        """
        return axis in self.actual_axes

    def index(self, axis: AxisType) -> int:
        """Index of ``axis`` in tensor i.e., returns  ``i`` if and only if ``axis`` is the ``i``-th
        dimension of the tensor.

        Args:
            axis (AxisType): a "named" axis

        Returns:
            int: the corresponding index
        """
        return self.actual_axes.index(axis)

    def reverse_index(self, axis: AxisType) -> int:
        """Index of ``axis`` in tensor but in reverse order i.e., returns  ``-i`` if and only if
        ``axis`` is the ``(n - i)``-th dimension of the tensor (with ``n`` the total number of dimension).
        dimension of the tensor.

        Args:
            axis (AxisType): a "named" axis

        Returns:
            int: the corresponding index in reverse order
        """
        return self.actual_axes.reverse_index(axis)

    def size(
        self,
        axis: AxisType,
    ) -> int:
        """The size of a given ``axis``. Returns 0 when the ``axis`` is "flattened".

        Args:
            axis (AxisType): a "named" axis

        Raises:
            ValueError: if ``axis`` is not a possible axis

        Returns:
            int: the size of ``axis`` (0 when the ``axis`` is "flattened")
        """
        if not self.has(axis):
            raise ValueError(
                f"{axis} is not a valid axis."
                f"Valid axes are: {', '.join(str(a) for a in self.axes)}"
            )
        if not self.has_actual(axis):
            return 0
        reverse_idx = self.reverse_index(axis)
        return self.shape[reverse_idx]

    def is_broadcastable_with(
        self,
        other_tensor_or_shape: Union[Tensor[ArrayType, AxisType], Tuple[int, ...]],
    ) -> bool:
        """Whether this tensor can be
        `broadcasted <https://numpy.org/doc/stable/user/basics.broadcasting.html>`_
        with ``other_tensor_or_shape``.

        Args:
            other_tensor_or_shape (Union[Tensor[ArrayType, AxisType], Tuple[int, ...]]): another
                tensor

        Returns:
            bool: ``True`` if and only if this tensor can be broadcasted with ``other_tensor_or_shape``,
                ``False`` otherwise.
        """
        return is_broadcastable(
            self.shape,
            other_tensor_or_shape
            if isinstance(other_tensor_or_shape, tuple)
            else other_tensor_or_shape.shape,
        )


class RegularizedArrayLikeCost(AverageableArrayLike[ArrayType], Protocol[ArrayType]):
    """Interface for "regularized" costs i.e., costs of the form:
    ``cost + lagrangian_coefficient * regularization``

    `More about regularization <https://en.wikipedia.org/wiki/Regularization_(mathematics)>`_.

    Args:
        cost (AverageableArrayLike[ArrayType]): an array-like collection of costs
        regularization (AverageableArrayLike[ArrayType]): an array-like collection regularization costs
        lagrangian_coefficient (Number): a non-negative number quantifying the regularization weight
    """

    cost: AverageableArrayLike[ArrayType]
    regularization: AverageableArrayLike[ArrayType]
    lagrangian_coefficient: Number = 1  # type: ignore[assignment]

    def mean(self, *args, **kwargs) -> Union[ArrayType, Any]:
        return (
            self.cost.mean(*args, **kwargs)
            + self.regularization.mean(*args, **kwargs) * self.lagrangian_coefficient
        )



"""Classes and methods to optimize functions involving jax arrays transformations via gradient descent"""

from __future__ import annotations

import functools
from collections import UserList, deque
from dataclasses import dataclass, replace
from typing import Deque, Generic, Hashable, List, Optional, Protocol, Tuple, TypeVar

import jax.numpy as jnp
import jax_dataclasses as jdc
import optax
from matplotlib import pyplot as plt
from tqdm.autonotebook import trange
from typing_extensions import Self

from jax_utils.compilation import BaseJaxCompilable, jit_when_compilation_enabled
from jax_utils.gradient import BaseGradientStep
from jax_utils.jax_tensor import AverageableJaxArrayLike, JaxTensor

State = TypeVar("State")
Action = TypeVar("Action")
Observation = TypeVar("Observation")
Cost = TypeVar("Cost", bound=AverageableJaxArrayLike)
JaxTensorType = TypeVar("JaxTensorType", bound=JaxTensor)
OptimizerState = TypeVar("OptimizerState", bound=optax.OptState)
AxisType = TypeVar("AxisType", bound=Hashable)


@dataclass(frozen=True)
class OptimizationState(Generic[State, Action, Cost, OptimizerState]):
    """The current state of an iterative optimization procedure involving the
    :class:`jax_utils.markov_decision_process.Dynamics`
    of a Markov Decision Process (typically a cost minimization).

    Args:
        iteration (int): the current iteration step
        state (State): the MDP state
        action (Action): the MDP current action
        cost (Optional[Cost], optional): the current cost associated to the state-action pair. Defaults to None.
        optimizer_state (Optional[OptimizerState], optional): The current ``optax.OptState``. Defaults to None.
    """

    iteration: int
    state: State
    action: Action
    cost: Optional[Cost] = None
    optimizer_state: Optional[OptimizerState] = None

    @property
    def scalar_cost(self) -> float:
        """Returns the average ``self.cost``

        Returns:
            float: average cost
        """
        return float("inf") if self.cost is None else float(self.cost.mean())


class OptimStoppingCondition(BaseJaxCompilable, Protocol[State, Action, Cost, OptimizerState]):
    """Interface for all stopping conditions of an iterative optimization procedure involving the
    :class:`jax_utils.markov_decision_process.Dynamics` of a Markov Decision Process (typically a cost minimization).

    The stopping condition depends on an ``OptimizationState`` provided as input but may
    also collect information over steps.
    """

    def stop(
        self, optimization_state: OptimizationState[State, Action, Cost, OptimizerState]
    ) -> bool:
        """Returns a boolean inidicating whether the optimization procedure should be stopped.
        This class should be overriven in every concrete class.

        Args:
            optimization_state (OptimizationState[State, Action, Cost, OptimizerState]): _description_

        Returns:
            bool: `True`` if the optimization procedure should be stopped, `False`` otherwise.
        """

    def reset(self) -> Self:
        """Resets the stopping condition to an initial configuration."""
        return self

    @property
    def nb_iterations_upper_bound(self) -> int:
        """An upper bound on the maximal number of iterations. Default is 1e20."""
        return int(1e20)

    def __and__(
        self, other_stopping_condition: OptimStoppingCondition[State, Action, Cost, OptimizerState]
    ) -> OptimStoppingConditionIntersection:
        """Constructs a stopping condition that stops if and only if both ``self`` AND ``other_stopping_condition``
        stop.

        Args:
            other_stopping_condition (OptimStoppingCondition[State, Action, Cost, OptimizerState]): an other
                stopping condition

        Returns:
            OptimStoppingConditionIntersection: intersection of two stopping conditions
        """
        return OptimStoppingConditionIntersection(
            stopping_condition_1=self, stopping_condition_2=other_stopping_condition
        )

    def __or__(
        self, other_stopping_condition: OptimStoppingCondition[State, Action, Cost, OptimizerState]
    ) -> OptimStoppingConditionUnion:
        """Constructs a stopping condition that stops if and only if any of ``self`` OR ``other_stopping_condition``
        stops.

        Args:
            other_stopping_condition (OptimStoppingCondition[State, Action, Cost, OptimizerState]): an other
                stopping condition

        Returns:
            OptimStoppingConditionUnion: union of two stopping conditions
        """
        return OptimStoppingConditionUnion(
            stopping_condition_1=self, stopping_condition_2=other_stopping_condition
        )


class OptimStoppingConditionsCombination(
    OptimStoppingCondition[State, Action, Cost, OptimizerState], Protocol
):
    """This interface represents the combination of 2 stopping conditions.

    Args:
        stopping_condition_1 (OptimStoppingCondition[State, Action, Cost, OptimizerState]): a stopping condition
        stopping_condition_2 (OptimStoppingCondition[State, Action, Cost, OptimizerState]): a stopping condition
    """

    stopping_condition_1: OptimStoppingCondition[State, Action, Cost, OptimizerState]
    stopping_condition_2: OptimStoppingCondition[State, Action, Cost, OptimizerState]

    def enable_compilation(self) -> Self:
        self.stopping_condition_1.enable_compilation()
        self.stopping_condition_2.enable_compilation()
        return super().enable_compilation()

    def disable_compilation(self) -> Self:
        self.stopping_condition_1.disable_compilation()
        self.stopping_condition_2.disable_compilation()
        return super().disable_compilation()


@dataclass(frozen=True)
class OptimStoppingConditionIntersection(
    OptimStoppingConditionsCombination[State, Action, Cost, OptimizerState]
):
    """A combination of 2 stopping conditions using an "AND" operation.

    Args:
        stopping_condition_1 (OptimStoppingCondition[State, Action, Cost, OptimizerState]): a stopping condition
        stopping_condition_2 (OptimStoppingCondition[State, Action, Cost, OptimizerState]): a stopping condition
    """

    stopping_condition_1: OptimStoppingCondition[State, Action, Cost, OptimizerState]
    stopping_condition_2: OptimStoppingCondition[State, Action, Cost, OptimizerState]

    def stop(
        self, optimization_state: OptimizationState[State, Action, Cost, OptimizerState]
    ) -> bool:
        return self.stopping_condition_1.stop(
            optimization_state
        ) and self.stopping_condition_2.stop(optimization_state)

    def reset(self) -> OptimStoppingConditionIntersection[State, Action, Cost, OptimizerState]:
        self.stopping_condition_1.reset()
        self.stopping_condition_2.reset()
        return self

    def __repr__(self) -> str:
        return f"({repr(self.stopping_condition_1)} & {repr(self.stopping_condition_2)})"

    @property
    def nb_iterations_upper_bound(self) -> int:
        return max(
            self.stopping_condition_1.nb_iterations_upper_bound,
            self.stopping_condition_2.nb_iterations_upper_bound,
        )


@dataclass(frozen=True)
class OptimStoppingConditionUnion(
    OptimStoppingConditionsCombination[State, Action, Cost, OptimizerState]
):
    """A combination of 2 stopping conditions using an "OR" operation.

    Args:
        stopping_condition_1 (OptimStoppingCondition[State, Action, Cost, OptimizerState]): a stopping condition
        stopping_condition_2 (OptimStoppingCondition[State, Action, Cost, OptimizerState]): a stopping condition
    """

    stopping_condition_1: OptimStoppingCondition[State, Action, Cost, OptimizerState]
    stopping_condition_2: OptimStoppingCondition[State, Action, Cost, OptimizerState]

    def stop(
        self, optimization_state: OptimizationState[State, Action, Cost, OptimizerState]
    ) -> bool:
        return self.stopping_condition_1.stop(optimization_state) or self.stopping_condition_2.stop(
            optimization_state
        )

    def reset(self) -> OptimStoppingConditionUnion[State, Action, Cost, OptimizerState]:
        self.stopping_condition_1.reset()
        self.stopping_condition_2.reset()
        return self

    def __repr__(self) -> str:
        return f"({repr(self.stopping_condition_1)} | {repr(self.stopping_condition_2)})"

    @property
    def nb_iterations_upper_bound(self) -> int:
        return min(
            self.stopping_condition_1.nb_iterations_upper_bound,
            self.stopping_condition_2.nb_iterations_upper_bound,
        )


@dataclass(frozen=True)
class MaxIterationsStoppingCondition(OptimStoppingCondition[State, Action, Cost, OptimizerState]):
    """A stopping condition that stops when the number of iterations exceeds a given threshold.

    Args:
        max_iterations (int): maximal number of iterations before the stopping condition is raised
    """

    max_iterations: int

    def stop(
        self, optimization_state: OptimizationState[State, Action, Cost, OptimizerState]
    ) -> bool:
        if optimization_state.iteration >= self.max_iterations:
            print(f"Maximum number of {self.max_iterations} iterations reached.")
            return True
        return False

    def __repr__(self) -> str:
        return f"MaxIterations({self.max_iterations})"

    @property
    def nb_iterations_upper_bound(self) -> int:
        return self.max_iterations


@dataclass(frozen=True)
class MinIterationsStoppingCondition(OptimStoppingCondition[State, Action, Cost, OptimizerState]):
    """A stopping condition that continues as long as the number of iterations is below a given threshold.

    Args:
        min_iterations (int): minimal number of iterations before the stopping condition is raised
    """

    min_iterations: int

    def stop(
        self, optimization_state: OptimizationState[State, Action, Cost, OptimizerState]
    ) -> bool:
        if optimization_state.iteration <= self.min_iterations:
            return False
        return True

    def __repr__(self) -> str:
        return f"MinIterations({self.min_iterations})"


@jdc.pytree_dataclass
class MinDeltaActionStoppingCondition(
    OptimStoppingCondition[State, JaxTensorType, Cost, OptimizerState]
):
    """Stops when the action stops significantly changing (as defined by relative & absolute tolerance).
    The previous action is saved in memory to compare it to the new action.

    The delta between previous and new action is then compared to absolute and relative tolerance i.e, the
    stopping condition is raised when:
    ``all(abs(new_action - previous_action) <= absolute tolerance + relative_tolerance * maximum(abs(new_action),
    abs(previous_action)))``

    Args:
        relative_tolerance (jnp.ndarray): relative tolerance for action variations. Default is 1e-6.
        absolute_tolerance (jnp.ndarray): absolute tolerance for action variations. Default is 1e-6.
    """

    relative_tolerance: jnp.ndarray = jnp.array(1e-6)
    absolute_tolerance: jnp.ndarray = jnp.array(1e-6)

    @jit_when_compilation_enabled()
    def _stop(self, previous_action: JaxTensorType, new_action: JaxTensorType) -> jnp.ndarray:
        delta_action = jnp.abs(new_action.values - previous_action.values)
        return jnp.all(
            delta_action
            <= self.absolute_tolerance
            + self.relative_tolerance
            * jnp.maximum(jnp.abs(previous_action.values), jnp.abs(new_action.values))
        )

    def stop(
        self, optimization_state: OptimizationState[State, JaxTensorType, Cost, OptimizerState]
    ) -> bool:
        new_action = optimization_state.action
        try:
            previous_action = getattr(self, "_previous_action")
            should_stop = self._stop(previous_action, new_action)
        except AttributeError:
            should_stop = False
        object.__setattr__(self, "_previous_action", new_action)
        if should_stop:
            print("Changes in action are below specified (absolute + relative) tolerance.")
        return should_stop

    def reset(self) -> MinDeltaActionStoppingCondition[State, JaxTensorType, Cost, OptimizerState]:
        if hasattr(self, "_previous_action"):
            object.__delattr__(self, "_previous_action")
        return self


@jdc.pytree_dataclass
class MinDeltaCostStoppingCondition(
    OptimStoppingCondition[State, Action, jnp.ndarray, OptimizerState]
):
    """Stops when the cost stops significantly decreasing (as defined by relative & absolute tolerance).

    The ``window_length`` last cost values are saved in memory and the stopping condition is raised when the delta
    between the min and max of these values is below a threshold i.e., when:
    ``all(maximum(last_window_length_costs) - minimum(last_window_length_costs) < absolute_tolerance +
    relative_tolerance * minimum(last_window_length_costs))``

    Args:
        relative_tolerance (jnp.ndarray): relative tolerance for cost variations. Default is 1e-6.
        absolute_tolerance (jnp.ndarray): absolute tolerance for cost variations. Default is 1e-6.
        window_length (int): maximal length of the queue storing the past history of costs (costs older than
            ``window_length`` time steps are discarded)
    """

    relative_tolerance: jnp.ndarray = jnp.array(1e-6)
    absolute_tolerance: jnp.ndarray = jnp.array(1e-6)
    window_length: int = 2

    @jit_when_compilation_enabled()
    def _stop(self, costs_queue: List[jnp.ndarray]) -> jnp.ndarray:
        max_cost = functools.reduce(lambda x, y: jnp.maximum(x, y), costs_queue)
        min_cost = functools.reduce(lambda x, y: jnp.minimum(x, y), costs_queue)
        delta_cost = max_cost - min_cost
        return jnp.all(
            delta_cost
            <= self.relative_tolerance * jnp.minimum(jnp.abs(max_cost), jnp.abs(min_cost))
            + self.absolute_tolerance
        )

    def stop(
        self, optimization_state: OptimizationState[State, Action, jnp.ndarray, OptimizerState]
    ) -> bool:
        try:
            if not hasattr(self, "_costs_queue"):
                return False
            costs_queue: Deque[jnp.ndarray] = getattr(self, "_costs_queue")
            if len(costs_queue) < self.window_length:
                return False
            should_stop = self._stop(list(costs_queue))
            if should_stop:
                print("Cost decrease is below specified (absolute + relative) tolerance.")
                return True
            return False
        finally:
            if not hasattr(self, "_costs_queue"):
                object.__setattr__(self, "_costs_queue", deque())
            costs_queue: Deque[jnp.ndarray] = getattr(self, "_costs_queue")  # type: ignore[no-redef]
            if optimization_state.cost is not None:
                costs_queue.append(optimization_state.cost)
            if len(costs_queue) > self.window_length:
                costs_queue.popleft()

    def reset(self) -> MinDeltaCostStoppingCondition[State, Action, OptimizerState]:
        if hasattr(self, "_costs_queue"):
            object.__delattr__(self, "_costs_queue")
        return self


class CostHistory(UserList[AverageableJaxArrayLike]):
    """List of costs (typically jax arrays)"""

    def scalar_costs(self) -> List[float]:
        """
        Returns:
            List[float]: list of cost means
        """
        return [float(cost.mean()) for cost in self.data]

    def plot_scalar_costs(self):
        """Plots cost history"""
        plt.plot(self.scalar_costs())
        plt.title("Cost evolution over time")
        plt.xlabel("Iterations")
        plt.ylabel("Cost")


@dataclass(frozen=True)
class GradientDescentOptimizationLoop(
    Generic[AxisType, State, Action, Cost, Observation, OptimizerState]
):
    """Gradient descent optimization procedure involving the :class:`jax_utils.markov_decision_process.Dynamics`
    of a Markov Decision Process. This class is a callable that recursively applies ``gradient_step`` until
    a ``stopping_condition`` is met.

    Args:
        gradient_step (BaseGradientStep[AxisType, State, Action, Cost, OptimizerState]): a gradient step
        stopping_condition (OptimStoppingCondition[State, Action, Cost, OptimizerState]): a stopping condition
    """

    gradient_step: BaseGradientStep[AxisType, State, Action, Cost, OptimizerState]
    stopping_condition: OptimStoppingCondition[State, Action, Cost, OptimizerState]

    def __call__(
        self,
        state: State,
        init_action: Action,
        iteration: int = 0,
        cost: Optional[Cost] = None,
        opt_state: Optional[OptimizerState] = None,
    ) -> Tuple[
        OptimizationState[State, Action, Cost, OptimizerState],
        OptimizationState[State, Action, Cost, OptimizerState],
        CostHistory,
    ]:
        """Iteratively minimizes the cost of the MDP dynamics (in some state) by performing gradient descent
        on the action.

        Args:
            state (State): state of the MDP (does not change during optimization)
            init_action (Action): initial action that will then be optimized
            iteration (int, optional): current  iteration. Defaults to 0.
            cost (Optional[Cost], optional): current cost. Defaults to None.
            opt_state (Optional[OptimizerState], optional): current optimization state. Defaults to None.

        Returns:
            Tuple[ OptimizationState[State, Action, Cost, OptimizerState], OptimizationState[State, Action, Cost, OptimizerState], CostHistory,]: the optimal `OptimizationState`, final `OptimizationState` and history o cost during the optimization iterations  # pylint: disable=line-too-long

        """
        # Initialization
        # --------------------
        cost_history = CostHistory()
        opt_state = (
            self.gradient_step.init_optimizer(init_action) if opt_state is None else opt_state
        )
        action = init_action

        optimization_state = OptimizationState[State, Action, Cost, OptimizerState](
            iteration=iteration, state=state, action=action, cost=cost, optimizer_state=opt_state
        )
        optimal_optimization_state = replace(optimization_state)

        # Optim loop
        # --------------------
        pbar = trange(self.stopping_condition.nb_iterations_upper_bound)
        for it in pbar:
            action, opt_state, cost_value = self.gradient_step(
                state,
                action,
                opt_state,
            )
            cost_history.append(cost_value)
            optimization_state = replace(
                optimal_optimization_state,
                iteration=it,
                action=action,
                cost=cost_value,
                optimizer_state=opt_state,
            )
            scalar_cost = optimization_state.scalar_cost
            pbar.set_postfix({"cost": scalar_cost})

            if scalar_cost < optimal_optimization_state.scalar_cost:
                optimal_optimization_state = optimization_state

            if self.stopping_condition.stop(optimization_state):
                break

        return optimal_optimization_state, optimization_state, cost_history

    def resume(
        self, optimization_state: OptimizationState[State, Action, Cost, OptimizerState]
    ) -> Tuple[
        OptimizationState[State, Action, Cost, OptimizerState],
        OptimizationState[State, Action, Cost, OptimizerState],
        CostHistory,
    ]:
        """Resume optimization loop from ``optimization_state`` (cold start).

        Args:
            optimization_state (OptimizationState[State, Action, Cost, OptimizerState]): state from which to resume the
                optimization procedure.

        Returns:
            OptimizationState[State, Action, Cost, OptimizerState], OptimizationState[State, Action, Cost,OptimizerState], CostHistory,]: same outputs as :meth:`GradientDescentOptimizationLoop.__call__`  # pylint: disable=line-too-long
        """
        return self(
            state=optimization_state.state,
            init_action=optimization_state.action,
            iteration=optimization_state.iteration,
            cost=optimization_state.cost,
            opt_state=optimization_state.optimizer_state,
        )



"""Wrappers around jax arrays"""

from numbers import Number
from typing import (
    AbstractSet,
    Generic,
    Hashable,
    Optional,
    Protocol,
    Tuple,
    TypeAlias,
    TypeVar,
    Union,
    runtime_checkable,
)

import jax.numpy as jnp
import jax_dataclasses as jdc
from jax import nn
from typing_extensions import Self

from jax_utils.compilation import BaseJaxCompilable, jit_when_compilation_enabled
from jax_utils.pytree import ConvertibleToAxes
from jax_utils.common_tensor import AverageableArrayLike, RegularizedArrayLikeCost, Tensor

AxisType = TypeVar("AxisType", bound=Hashable)


@runtime_checkable
class JaxTensor(Tensor[jnp.ndarray, AxisType], ConvertibleToAxes[AxisType], Protocol[AxisType]):
    """Interface representing a ``Tensor`` with a
    `JAX <https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.array.html>`_ ``array``.

    A ``JaxTensor`` is always "convertible to axes" (see
    :class:`jax_utils.pytree.ConvertibleToAxes` interface) using
    :meth:`jax_utils.pytree.ConvertibleToAxes.convert_to_axes` method. When a ``Tensor`` is
    "converted to axes", the  associated ``array`` attribute is no longer a JAX array but rather an integer indicating
    the dimension index of the specified ``axis`` (argument of method ``convert_to_axes``). If ``axis`` is
    flattened or not present, ``self.array`` is set to ``None`` when converted to axes.

    Converting a :class:`jax_utils.jax_tensor.JaxTensor` to axes is useful to determine the
    ``in_axes`` and ``out_axes`` arguments of
    `jax.vmap <https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html>`_ and
    `jax.pmap <https://jax.readthedocs.io/en/latest/_autosummary/jax.pmap.html>`_.

    Args:
        array (jnp.ndarray): a JAX array (or an optional integer when "converted to axes")
    """

    array: jnp.ndarray

    def __init__(self, array: jnp.ndarray, **kwargs):
        self.array = array
        super().__init__()

    def __post_init__(self):
        if isinstance(self.array, jnp.ndarray):
            # in case the class is instantiated as a PyTreeDef (e.g., for vmap),
            # self.array may be an int, None, ...
            super().__post_init__()

    def is_broadcastable_with(
        self,
        other_tensor_or_shape: Union[Tensor[jnp.ndarray, AxisType], Tuple[int, ...]],
    ) -> bool:
        if isinstance(self.array, jnp.ndarray):
            # in case the class is instantiated as a PyTreeDef (e.g., for vmap),
            # self.array may be an int, None, ...
            return super().is_broadcastable_with(other_tensor_or_shape)
        return True

    @classmethod
    def from_flattened_axes(
        cls, array: jnp.ndarray, flattened_axes: AbstractSet[AxisType], **kwargs
    ) -> Self:
        """
        Constructor when some axes are "flattened" in ``array``.

        This method simply adds missing dimensions accordingly.

        Args:
            array (jnp.ndarray): a JAX array
            flattened_axes (AbstractSet[AxisType]): the named axes that are "flattened"

        Returns:
            Self: an instance of class ``cls``
        """
        return cls(
            array=jnp.expand_dims(array, axis=cls._expand_dims_axis(flattened_axes)), **kwargs
        )

    def convert_to_axes(self, axis: Optional[AxisType]) -> Self:
        """Convert the :class:`jax_utils.jax_tensor.JaxTensor` object so that it can be passed as
        the ``in_axes`` or ``out_axes`` arguments of
        `jax.vmap <https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html>`_ and
        `jax.pmap <https://jax.readthedocs.io/en/latest/_autosummary/jax.pmap.html>`_.

        Args:
            axis (Optional[AxisType]): a "named" axis

        Returns:
            Self: same object as ``self`` but with ``array`` replaced by the index of ``axis`` i.e.,
                if ``axis`` corresponds to the ``i``-th dimension of ``array``, then ``array`` is replaced by ``i``,
                if ``axis`` is a flattened axis or simply not present, then ``array`` is replaced by ``None``
        """
        with jdc.copy_and_mutate(self, validate=False) as jax_tensor:
            if axis is not None and axis in self.actual_axes:
                jax_tensor.array = self.reverse_index(axis)  # type: ignore[assignment]
            else:
                jax_tensor.array = None  # type: ignore[assignment]
        return jax_tensor


class NonNegativeValues(Protocol):
    """Interface to be used in combination with :class:`jax_utils.jax_tensor.JaxTensor` in order
    to implement JAX tensors that are constrained to take non-negative ``values``.

    Args:
        array (jnp.ndarray): a JAX array
    """

    array: jnp.ndarray

    # pylint: disable=C0116
    @property
    def values(self) -> jnp.ndarray:
        return nn.elu(self.array - 1) + 1

    # pylint: disable=C0116
    def reverse_values(self, values: jnp.ndarray) -> jnp.ndarray:
        return jnp.where(values > 1, values, jnp.log(values) + 1)


class NonNegativeBudgetedValues(NonNegativeValues, Protocol):
    """Interface to be used in combination with :class:`jax_utils.jax_tensor.JaxTensor` in order to
    implement JAX tensors that are constrained to both take non-negative ``values`` and have their
    sum over the last axis bounded by  a maximal "budget" (``max_budget``).

    Args:
        array (jnp.ndarray): a JAX array
        max_budget (Number): a non-negative number representing the maximum value that
            the sum over the last axis can take
    """

    array: jnp.ndarray
    max_budget: Number

    @property
    def values(self) -> jnp.ndarray:
        return (
            self.max_budget
            * super().values
            / (super().values.sum(axis=-1)[..., jnp.newaxis] + 1e-6)
        )[..., :-1]

    def reverse_values(self, values: jnp.ndarray) -> jnp.ndarray:
        output = jnp.zeros(self.array.shape).at[..., :-1].set(values)
        output = output.at[..., -1].set(jnp.maximum(self.max_budget - values.sum(axis=-1), 1e-6))
        return super().reverse_values(output)


AverageableJaxArrayLike: TypeAlias = AverageableArrayLike[jnp.ndarray]

JaxCostType = TypeVar("JaxCostType", bound=AverageableJaxArrayLike)
JaxRegularizedCostType = TypeVar("JaxRegularizedCostType", bound=AverageableJaxArrayLike)


@jdc.pytree_dataclass(frozen=True)
class RegularizedJaxCost(
    RegularizedArrayLikeCost[jnp.ndarray],
    BaseJaxCompilable,
    Generic[JaxCostType, JaxRegularizedCostType],
):
    """Interface for "regularized" JAX costs i.e., costs of the form:
    ``cost + lagrangian_coefficient * regularization``

    Args:
        cost (JaxCostType): an JAX array of costs
        regularization (JaxRegularizedCostType): a JAX array of regularization costs
        lagrangian_coefficient (jdc.Static[Number]): a non-negative number quantifying the
            regularization weight
    """

    cost: JaxCostType
    regularization: JaxRegularizedCostType
    lagrangian_coefficient: jdc.Static[Number] = 1  # type: ignore[assignment]

    @jit_when_compilation_enabled()
    def mean(self, *args, **kwargs) -> jnp.ndarray:
        return super().mean(*args, **kwargs)



"""Define useful types for package"""

from dataclasses import Field, dataclass
from typing import Any, ClassVar, Hashable, List, Optional, Protocol, Tuple, Union


class DataclassInstance(Protocol):
    """Type for Python dataclasses (see ``dataclasses.dataclass``).

    This code is copy-pasted from ``_typeshed`` (but is simpler to import).
    """

    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


class HashableIndexingOrSlicing(DataclassInstance, Hashable, Protocol):
    """Python slices and lists (of indices) are not hashable in Python (slices are hashable for Python version >= 3.12).
    This is a shared interface for hashable slices and indices.
    """

    @property
    def values(self) -> Union[slice, int, List[int]]:
        """
        Property returning index values in the form of a slice, an int (for singletons) or a list of int.

        Returns:
            Union[slice, int, List[int]]: index values
        """


@dataclass(eq=True, frozen=True)
class HashableSlicing(HashableIndexingOrSlicing):
    """Python slices are not hashable for Python version < 3.12.
    This class allows to define hashable slices.

    Args:
        start (Optional[int]): initial value of the slice (included)
        stop (Optional[int]): last value of the slice (excluded)
        step (Optional[int]): step between values
    """

    start: Optional[int] = None
    stop: Optional[int] = None
    step: Optional[int] = None

    @property
    def values(self) -> slice:
        return slice(self.start, self.stop, self.step)

    def __hash__(self):
        return hash((self.start, self.stop, self.step))


@dataclass(eq=True, frozen=True, init=False)
class HashableIndexing(HashableIndexingOrSlicing):
    """Python lists are not hashable.
    This class allows to define hashable list of indices.

    Args:
        indices (Union[int, Tuple[int]]): index or list of (integer-valued) indices
    """

    indices: Union[int, Tuple[int]]

    def __init__(self, *indices: int):
        object.__setattr__(self, "indices", tuple(indices) if len(indices) > 1 else indices[0])

    @property
    def values(self) -> Union[int, List[int]]:
        return self.indices if isinstance(self.indices, int) else list(self.indices)

    def __hash__(self):
        return hash(self.indices)



"""High-level abstractions for decision problems (Markov Decision processes, etc...)
involving jax arrays transformations"""

from typing import Hashable, Protocol, Tuple, TypeVar

import jax.numpy as jnp
import jax_dataclasses as jdc
import optax
from jax import value_and_grad

from jax_utils.compilation import jit_when_compilation_enabled
from jax_utils.jax_tensor import AverageableJaxArrayLike
from jax_utils.vectorization import JaxDataclassNestedConvertibleToAxes, vectorize
from jax_utils.markov_decision_process import CostRegularizer, Dynamics, RegularizedDynamics

State = TypeVar("State")
State_contra = TypeVar("State_contra", contravariant=True)
Action_contra = TypeVar("Action_contra", contravariant=True)
Action = TypeVar("Action")
Observation = TypeVar("Observation")
Observation_co = TypeVar("Observation_co", covariant=True)
Cost = TypeVar("Cost", bound=AverageableJaxArrayLike)
Cost_co = TypeVar("Cost_co", covariant=True, bound=AverageableJaxArrayLike)
RegularizedCost = TypeVar("RegularizedCost")
OptimizerState = TypeVar("OptimizerState", bound=optax.OptState)
GradientUpdates = TypeVar("GradientUpdates", bound=optax.Updates)
AxisType = TypeVar("AxisType", bound=Hashable)
AxisType_contra = TypeVar("AxisType_contra", contravariant=True, bound=Hashable)


class JaxDynamics(
    JaxDataclassNestedConvertibleToAxes[AxisType_contra],
    Dynamics[State, Action_contra, Cost_co, Observation_co],
    Protocol[AxisType_contra, State, Action_contra, Cost_co, Observation_co],
):
    """A :class:`jax_utils.markov_decision_process.Dynamics` involving Jax arrays transformations. The cost of such
    dynamics can be differentiated w.r.t. the action (see method
    :meth:`jax_utils.dynamics.JaxDynamics.compute_gradient`).
    """

    @jit_when_compilation_enabled()
    def scalar_cost(self, state: State, action: Action_contra) -> jnp.ndarray:
        """Averages the cost associated to the dynamics of an MDP.

        Args:
            state (State): a state of the MDP
            action (Action_contra): an action of the MDP

        Returns:
            jnp.ndarray: a scalar Jax array corresponding to the mean cost
        """
        return self(state, action)[1].mean()

    @jit_when_compilation_enabled()
    def compute_gradient(
        self, state: State, action: Action_contra
    ) -> Tuple[Cost_co, Action_contra]:
        """The cost of a :class:`jax_utils.dynamics.JaxDynamics` can always be differentiated
        w.r.t. the action.

        Args:
            state (State): a state of the MDP
            action (Action_contra): an action of the MDP

        Returns:
            Tuple[Cost_co, Action_contra]: the associated cost and the gradient of the action
        """
        return value_and_grad(self.scalar_cost, argnums=1)(state, action)


@jdc.pytree_dataclass(frozen=True)
class VectorizedJaxDynamics(JaxDynamics[AxisType, State, Action_contra, Cost_co, Observation_co]):
    """A :class:`jax_utils.dynamics.JaxDynamics` that can be
    `vectorized <https://jax.readthedocs.io/en/latest/jax-101/03-vectorization.html>`_ over some "named" axis.

    Args:
        dynamics (JaxDynamics[AxisType, State, Action_contra, Cost_co, Observation_co]): the
            :class:`jax_utils.dynamics.JaxDynamics` to be vectorized
        vectorized_axis (jdc.Static[AxisType]): "named" axis to map over
    """

    dynamics: JaxDynamics[AxisType, State, Action_contra, Cost_co, Observation_co]
    vectorized_axis: jdc.Static[AxisType]

    @vectorize()
    def __call__(
        self, state: State, action: Action_contra
    ) -> Tuple[State, Cost_co, Observation_co]:
        return self.dynamics(state, action)


@jdc.pytree_dataclass(frozen=True)
class RegularizedJaxDynamics(
    RegularizedDynamics[State, Action, Cost, RegularizedCost, Observation],
    JaxDynamics[AxisType_contra, State, Action, Cost, Observation],
):
    """A :class:`jax_utils.dynamics.JaxDynamics` with cost regularization.
    See also :class:`jax_utils.markov_decision_process.RegularizedDynamics`.

    Args:
        dynamics (JaxDynamics[AxisType_contra, State, Action, Cost, Observation]): the
            :class:`jax_utils.dynamics.JaxDynamics` to be vectorized
        cost_regularizer (CostRegularizer[State, Action, Cost, Observation, RegularizedCost]): a cost regularizer
    """

    dynamics: JaxDynamics[AxisType_contra, State, Action, Cost, Observation]
    cost_regularizer: CostRegularizer[State, Action, Cost, Observation, RegularizedCost]

    @jit_when_compilation_enabled()
    def __call__(self, state: State, action: Action) -> Tuple[State, RegularizedCost, Observation]:
        return super().__call__(state, action)



"""Jax arrays transformations (scaling, ...)"""

from __future__ import annotations

from typing import Any, Generic, Hashable, Set, Tuple, Type, TypeVar

import jax.numpy as jnp
import jax_dataclasses as jdc
from jax.tree_util import tree_map

from jax_utils.compilation import BaseJaxCompilable, jit_when_compilation_enabled
from jax_utils.jax_tensor import JaxTensor
from jax_utils.common_tensor import TensorAxes, expand_dims_axis

AxisType = TypeVar("AxisType", bound=Hashable)


def scale_jax_tensor(
    tensor: JaxTensor[AxisType], scaling_factor: jnp.ndarray
) -> JaxTensor[AxisType]:
    """Scale the ``values`` of a ``JaxTensor`` i.e., multiply ``values`` by ``scaling_factor``.
    Note that property ``values`` in ``JaxTensor`` differ from attribute ``array``.

    Args:
        tensor (JaxTensor[AxisType]): jax tensor to be scaled
        scaling_factor (jnp.ndarray): the scaling factors to apply for scaling

    Returns:
        JaxTensor[AxisType]: scaled jax tensor
    """
    with jdc.copy_and_mutate(tensor, validate=True) as scaled_tensor:
        scaled_tensor.array = scaled_tensor.reverse_values(scaled_tensor.values * scaling_factor)
    return scaled_tensor


def unscale_jax_tensor(
    scaled_tensor: JaxTensor[AxisType], scaling_factor: jnp.ndarray
) -> JaxTensor[AxisType]:
    """Reverse transformation of function ``scale_jax_tensor``.

    Args:
        tensor (JaxTensor[AxisType]): jax tensor to be unscaled
        scaling_factor (jnp.ndarray): the scaling factors applied when scaling the original jax tensor

    Returns:
        JaxTensor[AxisType]: unscaled jax tensor
    """
    with jdc.copy_and_mutate(scaled_tensor, validate=True) as tensor:
        tensor.array = scaled_tensor.reverse_values(tensor.values / scaling_factor)
    return tensor


@jdc.pytree_dataclass(frozen=True)
class JaxScaler(BaseJaxCompilable, Generic[AxisType]):
    """When applying gradient descent optimization algorithms, it is often helpful to scale all the arrays/tensors
    involved in the computations.

    This class allows to easily scale/unscale :class:`jax_utils.jax_tensor.JaxTensor`'s (even when
    the tensors are stored in a nested pytree structure).

    Args:
        scaling_factors (jnp.ndarray): Multiplicative factors for scaling. Default to ``jnp.array(1.0)``.
        scaling_axes (jdc.Static[Set[AxisType]]): "Named" axes of ``scaling_factors``. Thus, the number of axes should
            match ``scaling_factors.ndim``. Default to ``jdc.field(default_factory=set)``.
        tensor_types_to_scale (jdc.Static[Tuple[Type[JaxTensor]]]): Object types that should be scaled in a pytree.
            Default to :class:`jax_utils.jax_tensor.JaxTensor`.
        tensor_types_not_to_scale (jdc.Static[Tuple[JaxTensor]]): Object types that should not be scaled in a pytree.
            Default to ``tuple()``.
    """

    scaling_factors: jnp.ndarray = jnp.array(1.0)
    scaling_axes: jdc.Static[Set[AxisType]] = jdc.field(default_factory=set)
    tensor_types_to_scale: jdc.Static[Tuple[Type[JaxTensor]]] = (JaxTensor,)
    tensor_types_not_to_scale: jdc.Static[Tuple[JaxTensor]] = tuple()  # type: ignore[assignment]

    def __post_init__(self):
        if self.scaling_factors.ndim != len(self.scaling_axes):
            raise ValueError(
                "the number of scaling_axes should math the number of dimensions of scaling_factors"
            )
        if len(set(self.tensor_types_to_scale) & set(self.tensor_types_not_to_scale)) > 0:
            raise ValueError(
                "tensor_types_to_scale and tensor_types_not_to_scale should be disjoint"
            )

    @classmethod
    def from_tensor(
        cls,
        tensor: JaxTensor[AxisType],
        scaling_axes: Set[AxisType],
        tensor_types_to_scale: jdc.Static[Tuple[Type[JaxTensor]]] = (JaxTensor,),
        tensor_types_not_to_scale: jdc.Static[Tuple[JaxTensor]] = tuple(),  # type: ignore[assignment]
        factor: float = 1.0,
    ) -> JaxScaler:
        """Alternative constructor that automatically computes the ``scaling_factors`` based on a
        :class:`jax_utils.jax_tensor.JaxTensor`, the ``scaling_axes``  and a target ``factor``.

        If ``my_scaler = JaxScaler.from_tensor(tensor=my_tensor, scaling_axes=my_axes, factor=my_factor)`` then
        ``my_scaler.scale(my_tensor)`` will be a :class:`jax_utils.jax_tensor.JaxTensor` with
        ``my_tensor.mean_over_axes(axes=my_axes)`` being equal to a "constant" Jax array with all elements equal to
        ``my_factor``.

        Args:
            tensor (JaxTensor[AxisType]): a jax tensor
            scaling_axes (Set[AxisType]): a set of "named" axes present in `tensor`
            tensor_types_to_scale (jdc.Static[Tuple[Type[JaxTensor]]]): Object types that should be scaled in a pytree.
                Default to :class:`jax_utils.jax_tensor.JaxTensor`.
            tensor_types_not_to_scale (jdc.Static[Tuple[JaxTensor]]): Object types that should not be scaled in a
                pytree. Default to ``tuple()``.

        Returns:
            JaxScaler: a scaler instance

        """
        axes_to_mean_over = tensor.actual_axes - scaling_axes
        return cls(
            scaling_factors=factor / jnp.maximum(tensor.mean_over_axes(axes=axes_to_mean_over), 1),
            scaling_axes=scaling_axes,
            tensor_types_to_scale=tensor_types_to_scale,
            tensor_types_not_to_scale=tensor_types_not_to_scale,
        )

    @jit_when_compilation_enabled()
    def scale(self, pytree: Any) -> Any:
        """Scales any pytree containing :class:`jax_utils.jax_tensor.JaxTensor`'s

        Args:
            pytree (Any): pytree to be scaled

        Returns:
            Any: the scaled pytree
        """
        return tree_map(
            lambda x: scale_jax_tensor(
                x,
                self.expand_scaling_factors(
                    tensor_axes=x.actual_axes, missing_axes=x.actual_axes - self.scaling_axes
                ),
            )
            if (
                isinstance(x, self.tensor_types_to_scale)
                and not isinstance(x, self.tensor_types_not_to_scale)  # type: ignore[arg-type]
            )
            else x,
            pytree,
            is_leaf=lambda x: isinstance(x, JaxTensor),
        )

    @jit_when_compilation_enabled()
    def unscale(self, scaled_pytree: Any) -> Any:
        """Unscales any pytree containing :class:`jax_utils.jax_tensor.JaxTensor`'s (reverse
        transformation of method ``scale``).

        Args:
            scaled_pytree (Any): pytree to be unscaled

        Returns:
            Any: the unscaled pytree
        """
        return tree_map(
            lambda x: unscale_jax_tensor(
                x,
                self.expand_scaling_factors(
                    tensor_axes=x.actual_axes, missing_axes=x.actual_axes - self.scaling_axes
                ),
            )
            if (
                isinstance(x, self.tensor_types_to_scale)
                and not isinstance(x, self.tensor_types_not_to_scale)  # type: ignore[arg-type]
            )
            else x,
            scaled_pytree,
            is_leaf=lambda x: isinstance(x, JaxTensor),
        )

    @jit_when_compilation_enabled()
    def expand_scaling_factors(
        self, tensor_axes: TensorAxes[AxisType], missing_axes: Set[AxisType]
    ) -> jnp.ndarray:
        """Expands the dimensions of ``self.scaling_factors`` to match the ``tensor_axes`` given as inputs.

        This method only adds new dimensions of size 1.

        Args:
            tensor_axes (TensorAxes[AxisType]): a set of "named" axes
            missing_axes (Set[AxisType]):

        Returns:
            jnp.ndarray: same as ``self.scaling_factors`` but with (empty) additional dimensions.
                The total number of dimensions of the returned array is equal to ``len(tensor_axes)``.
        """
        return jnp.expand_dims(
            self.scaling_factors,
            axis=expand_dims_axis(tensor_axes=tensor_axes, missing_axes=missing_axes),
        )



"""Classes and methods to easily vectorize jax tensors"""

from __future__ import annotations

from functools import wraps
from typing import Callable, Hashable, Optional, Protocol, TypeVar, runtime_checkable

import jax_dataclasses as jdc
import optax
from jax import vmap
from typing_extensions import Self

from jax_utils.compilation import (
    BaseJaxCompilable,
    JaxCompilableProtocol,
    jit_when_compilation_enabled,
)
from jax_utils.pytree import ConvertibleToAxes, pytree_to_axes
from jax_utils.jax_tensor import AverageableJaxArrayLike
from jax_utils.typing import DataclassInstance

State = TypeVar("State")
Action_contra = TypeVar("Action_contra", contravariant=True)
Action = TypeVar("Action")
Observation_co = TypeVar("Observation_co", covariant=True)
Cost_co = TypeVar("Cost_co", covariant=True, bound=AverageableJaxArrayLike)
OptimizerState = TypeVar("OptimizerState", bound=optax.OptState)
AxisType = TypeVar("AxisType", bound=Hashable)
AxisType_contra = TypeVar("AxisType_contra", contravariant=True, bound=Hashable)


@runtime_checkable
class JaxVectorizableProtocol(JaxCompilableProtocol, ConvertibleToAxes[AxisType], Protocol):
    """Interface for classes manipulating JAX arrays. It defines a special attribute ``vectorized_axis``
    corresponding to the "named" axis over which to apply
    `vectorization <https://jax.readthedocs.io/en/latest/jax-101/03-vectorization.html>`_.

    "Vectorizable" classes are also "compilable" (see interface
    :class:`jax_utils.compilation.JaxCompilableProtocol`) and convertible to axes (see
    :class:`jax_utils.pytree.ConvertibleToAxes` interface).

    Args:
        vectorized_axis (AxisType): "named" axis to be vectorized
    """

    vectorized_axis: AxisType


def vectorize(
    in_default_axis: Optional[int] = None,
    out_default_axis: Optional[int] = -1,
) -> Callable[[Callable], Callable]:
    """Parametrized decorator for methods of classes implementing :class:`JaxVectorizableProtocol`
    interface. It allows to
    `vectorize <https://jax.readthedocs.io/en/latest/jax-101/03-vectorization.html>`_
    the decorated methods according to a specified axis, namely attribute
    :attr:`JaxVectorizableProtocol.vectorized_axis` (see :class:`JaxVectorizableProtocol`).

    The `pytrees <https://jax.readthedocs.io/en/latest/pytrees.html>`_ of all inputs and outputs of the decorated
    method are explored and every JAX array as well as every class implementing the
    :class:`jax_utils.pytree.ConvertibleToAxes` interface is considered a leaf of the pytree.
    When a :class:`jax_utils.pytree.ConvertibleToAxes` leaf  is encountered, method
    :meth:`jax_utils.pytree.ConvertibleToAxes.convert_to_axes`
    is used to determine the dimension to vectorize. When a JAX array leaf or any other leaf is encountered instead,
    the default value ``in_default_axis`` (for input arguments) or ``out_default_axis`` (for output arguments) is used
    to determine the dimension to vectorize.

    Args:
        in_default_axis (Optional[int], optional): default index used in ``in_axes`` argument of method
            `jax.vmap <https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html>`_
            when a leaf that is not :class:`jax_utils.pytree.ConvertibleToAxes` is encountered
            in an input pytree. Defaults to None.
        out_default_axis (Optional[int], optional): default index used in ``out_axes`` argument of method
            `jax.vmap <https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html>`_ when a leaf that is not
            :class:`jax_utils.pytree.ConvertibleToAxes` is encountered in an output pytree.
            Defaults to -1.

    Returns:
        Callable[[Callable], Callable]: decorator with parameters ``in_default_axis`` and ``out_default_axis`` specified
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(self: JaxVectorizableProtocol, *args, **kwargs):
            if not hasattr(self, "_vmap_functions"):
                object.__setattr__(self, "_vmap_functions", {})

            vmap_functions = getattr(self, "_vmap_functions")

            in_axes = pytree_to_axes(
                (self,) + tuple(args) + tuple(kwargs.values()),
                self.vectorized_axis,
                default_axis=in_default_axis,
            )

            if in_axes not in vmap_functions:
                outputs = func(self, *args, **kwargs)
                out_axes = pytree_to_axes(
                    outputs, self.vectorized_axis, default_axis=out_default_axis
                )

                vmap_functions[in_axes] = vmap(func, in_axes=in_axes, out_axes=out_axes)

            return vmap_functions[in_axes](self, *args, **kwargs)

        return jit_when_compilation_enabled()(wrapper)

    return decorator


@runtime_checkable
class JaxDataclassNestedConvertibleToAxes(
    DataclassInstance,
    ConvertibleToAxes[AxisType_contra],
    BaseJaxCompilable,
    Protocol[AxisType_contra],
):
    """Interface for dataclasses that can be jit-compiled and are "convertible to axes", with a concrete implementation
    of ``convert_to_axes`` method that is specific to dataclasses.
    """

    def convert_to_axes(self, axis: Optional[AxisType_contra]) -> Self:
        """Concrete implementation of ``convert_to_axes`` for dataclasses.
        All the fields of the dataclass are inspected and method ``convert_to_axes``
        is applied whenever the field is ``ConvertibleToAxes``.

        Args:
            axis (Optional[AxisType_contra]): a "named" axis

        Returns:
            Self: same object as ``self`` but with array-like fields converted to axes
        """
        with jdc.copy_and_mutate(self, validate=False) as dc:
            for field in jdc.fields(self):
                field_value = getattr(self, field.name)
                if isinstance(field_value, ConvertibleToAxes):
                    setattr(dc, field.name, field_value.convert_to_axes(axis))
        return dc
